---------------------------------------------------------------------
--                    Open_Image - Test PPM
--
--                     Copyright (C) 2000
--                     Jeffrey Creem
--
-- This library is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public
-- License as published by the Free Software Foundation; either
-- version 2 of the License, or (at your option) any later version.
--
-- This library is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-- General Public License for more details.
--
-- You should have received a copy of the GNU General Public
-- License along with this library; if not, write to the
-- Free Software Foundation, Inc., 59 Temple Place - Suite 330,
-- Boston, MA 02111-1307, USA.
--
-- As a special exception, if other files instantiate generics from
-- this unit, or you link this unit with other files to produce an
-- executable, this  unit  does not  by itself cause  the resulting
-- executable to be covered by the GNU General Public License. This
-- exception does not however invalidate any other reasons why the
-- executable file  might be covered by the  GNU Public License.
---------------------------------------------------------------------

--  <description>
--
--  Simple test driver for testing the open_image.ppm_io package.
--
--  </description>
--
-- <notes>
--
-- This test driver is certainly not as clean as I would like nor
-- does it get 100% coverage of the package. Since I am sort of treating
-- this initial release like a detailed design release I am moving on
-- to a few new features to test out the overall feel of the API that
-- open_image supports. Any help at cleaning up this driver would be great.
--
-- </notes>
with Open_Image;
with Open_Image.Ppm_Io;
with Unchecked_Deallocation;
with Text_Io;

procedure Test_Ppm is

   Not_File : Text_Io.File_Type;


   type Pixel_Data is array (Open_Image.Column_Index_Type range <>, Open_Image.Row_Index_Type range <>) of Open_Image.Color_Channels_Type;

   type Access_Pixel_Data is access Pixel_Data;

   My_Raw_Image   : Access_Pixel_Data;
   My_Ascii_Image : Access_Pixel_Data;
   External_Ppm   : Access_Pixel_Data;


   procedure Free is
   new Unchecked_Deallocation(Pixel_Data, Access_Pixel_Data);


   function Get_Pixel (
         Image  : in     Integer;
         Column : in     Open_Image.Column_Index_Type;
         Row    : in     Open_Image.Row_Index_Type     )
     return Open_Image.Color_Channels_Type;

   pragma Inline(Get_Pixel);

   function Get_Pixel (
         Image  : in     Integer;
         Column : in     Open_Image.Column_Index_Type;
         Row    : in     Open_Image.Row_Index_Type     )
     return Open_Image.Color_Channels_Type is

      use Open_Image;

      Blue_Value : Open_Image.Pixel_Value_Type;

   begin

      Blue_Value := Pixel_Value_Type(Column);


      if (Column / 2) * 2 = Column then
         return (
            Red   => 128,
            Green => 0,
            Blue  => Blue_Value);
      else
         return (
            Red   => 0,
            Green => 128,
            Blue  => Blue_Value);
      end if;
   end Get_Pixel;

   procedure Create_Ppm is
   new Open_Image.Ppm_Io.Write_Ppm
      (
      Image_Type => Integer,
      Get_Pixel  => Get_Pixel);


   procedure Set_Pixel (
         Img  : in out Access_Pixel_Data;
         Column : in     Open_Image.Column_Index_Type;
         Row    : in     Open_Image.Row_Index_Type;
         Value  : in     Open_Image.Color_Channels_Type );
   pragma Inline(Set_Pixel);


   procedure Set_Size (
         Img     : in out Access_Pixel_Data;
         Max_Color : in     Open_Image.Pixel_Value_Type;
         Columns   : in     Open_Image.Column_Index_Type;
         Rows      : in     Open_Image.Row_Index_Type     );


   procedure Set_Pixel (
         Img  : in out Access_Pixel_Data;
         Column : in     Open_Image.Column_Index_Type;
         Row    : in     Open_Image.Row_Index_Type;
         Value  : in     Open_Image.Color_Channels_Type ) is


   begin
      Img.all(Column,Row) := Value;
   end Set_Pixel;

   procedure Read_Image is
   new Open_Image.Ppm_Io.Read_Ppm(
      Image_Type => Access_Pixel_Data,
      Set_Size   => Set_Size,
      Set_Pixel  => Set_Pixel);


   procedure Set_Size (
         Img     : in out Access_Pixel_Data;
         Max_Color : in     Open_Image.Pixel_Value_Type;
         Columns   : in     Open_Image.Column_Index_Type;
         Rows      : in     Open_Image.Row_Index_Type     ) is

      use type Open_Image.Column_Index_Type;
      use type Open_Image.Row_Index_Type;

   begin
      Free(Img);
      Image := new Pixel_Data(1 .. Columns, 1 .. Rows);
   end Set_Size;



begin
   Text_Io.Put_Line("Creating a rawbits image");

   Create_Ppm(
      Name                  => "test_rawbits.ppm",
      Max_Pixel_Value       => 255,
      Columns               => 128,
      Rows                  => 256,
      Store_Pixel_Values_As => Open_Image.Ppm_Io.Raw_Bits,
      Image                 => 2);

   Text_Io.Put_Line("Creating an all ASCII image");

   Create_Ppm(
      Name                  => "test_ascii.ppm",
      Max_Pixel_Value       => 255,
      Columns               => 128,
      Rows                  => 256,
      Store_Pixel_Values_As => Open_Image.Ppm_Io.Ascii_Data,
      Image                 => 2);

   Text_Io.Put_Line("Reading ASCII Image");

   Read_Image(
      Name  => "test_ascii.ppm",
      Image => My_Ascii_Image);

   Text_Io.Put_Line("Reading rawbits image");

   Read_Image(
      Name  => "test_rawbits.ppm",
      Image => My_Raw_Image);


   Text_Io.Put_Line("Doing sanity checks on written data v.s. read data");

   if My_Ascii_Image.all /= My_Raw_Image.all then

      Text_Io.Put_Line("*** Data from ASCII version and binary version");
      Text_Io.Put_Line("*** do not match and they should");

   end if;

   declare
      Data_Mismatch : Boolean := False;
      use type Open_Image.Color_Channels_Type;

   begin

      for Col in My_Raw_Image.all'range(1) loop
         for Row in My_Raw_Image.all'range(2) loop
            if My_Raw_Image(Col, Row) /= Get_Pixel(2, Col, Row) then

               Data_Mismatch := True;

            end if;
         end loop;

      end loop;

      if Data_Mismatch then
         Text_Io.Put_Line(
            "*** Read PPM file did not match written PPM file");
      end if;

   end;



   --
   -- Ok, we made it this far. Now for a few error cases
   --
   Text_Io.Create(
      Name => "not_ppm.ppm",
      File => Not_File);
   Text_Io.Put_Line(
      Item => "Hello",
      File => Not_File);
   Text_Io.Close(Not_File);

   if Open_Image.Ppm_Io.Is_A_Ppm_File("not_ppm.ppm") then
      Text_Io.Put_Line("*** Test says non PPM file is a PPM File");
   end if;


   if not Open_Image.Ppm_Io.Is_A_Ppm_File("test_rawbits.ppm") then
      Text_Io.Put_Line("*** Is_A_PPM failed on rawbits");
   end if;


   if not Open_Image.Ppm_Io.Is_A_Ppm_File("test_ascii.ppm") then
      Text_Io.Put_Line("*** Is_A_PPM failed on ascii");
   end if;

   begin

      Read_Image(
         Name  => "not_ppm.ppm",
         Image => My_Raw_Image);

      Text_Io.Put_Line("*** That is very odd. We were able to read a non");
      Text_Io.Put_Line("*** PPM file with a PPM read routine.");

   exception

      when Open_Image.Ppm_Io.Not_A_Ppm_File_Error =>
         null; -- Good thats what we wanted.

      when others =>
         Text_Io.Put_Line(
            "*** Wrong exception raised in reading non PPM file");
   end;


   begin

      Create_Ppm(
         Name                  => "test_rule_violate.ppm",
         Max_Pixel_Value       => 256,
         Columns               => 128,
         Rows                  => 256,
         Store_Pixel_Values_As => Open_Image.Ppm_Io.Raw_Bits,
         Image                 => 2);


      Text_Io.Put_Line(
         "*** Hmm..That is bad. It let us create rawbits for data > 255");

   exception


      when Open_Image.File_Format_Protocol_Violation_Error =>
         null; -- Good

      when others =>
         Text_Io.Put_Line(
            "*** Hmm..wrong exception for rawbits with data > 8 bits");

   end;


   begin
      Text_Io.Put_Line("Reading externally generated PPM file");

      Read_Image(
         Name  => "open_image.ppm",
         Image => External_Ppm);

   exception
      when others =>
         Text_Io.Put_Line("Got error reading externally generated PPM");
   end;


   Text_Io.Put_Line(
      "Test complete. If you did not see any *** output then");
   Text_Io.Put_Line(
      "things are basically as ok as they were when I ran it");

end Test_Ppm;
